use std::collections::BTreeMap;
use actix::prelude::*;
use actix_files as fs;
use actix_web::{middleware, web, App, HttpServer};

mod game;
mod datasource;
mod config;
mod server {
    pub mod server;
    pub mod messages;
    pub mod gamelobby;
    pub mod protocol;
}
use crate::server::server::{Server, ws_initiate};


#[actix_web::main]
async fn main() -> std::io::Result<()> {

    let args = clap::App::new("Acorn server")
        .version("0.1.0")
        .author("Nicolas Winkler <nicolas.winkler@gmx.ch>")
        .about("Server to play Elchhäutgerät games")
        .arg(clap::Arg::with_name("config")
                 .short("c")
                 .long("config")
                 .takes_value(true)
                 .help("config file"))
        .get_matches();
    
    let config_path = args.value_of("config").unwrap_or("config.toml");
    let config_txt = std::fs::read_to_string(config_path).expect("could not open config file");
    let config = config::parse_config(&config_txt).expect("invalid config file");

    let server_bind_addr = config.server_config.ip + ":" +
        &match config.server_config.port {
            Some(p) => p.to_string(),
            None => "8000".to_owned()
        };

    //std::env::set_var("RUST_LOG", "actix_server=info,actix_web=info");
    env_logger::init();

    let server = Server {
        lobbies: BTreeMap::new(),
        default_data: datasource::create_array_source(&config.default_questions)
    };
    let server_addr = server.start();

    let server = HttpServer::new(move || {
        App::new()
            .app_data(web::Data::new(server_addr.clone()))
            .wrap(middleware::Logger::default())
            .service(web::resource("/ws/").route(web::get().to(ws_initiate)))
            .service(fs::Files::new("/", "static/").index_file("index.html"))
    })
    .bind(server_bind_addr)?;

    let server = if let Some(t) = config.server_config.workers {
        server.workers(t as usize)
    } else {
        server
    };

    server.run().await
}
